from math import atan2, radians, degrees, sin, cos
from random import random, randint, choice

import pyglet
from pyglet.window import key

from cocos.actions import CallFunc, Delay
from cocos.collision_model import CollisionManagerGrid
from cocos.director import director
from cocos.euclid import Vector2
from cocos.layer import Layer, ColorLayer
from cocos.text import Label

from shrooms import cfg
from shrooms.sounds import sounds
from shrooms.sprites import (Background, Player, Moon, Mushroom, Bullet,
                             Explosion, Earth, Sun, Star)


pyglet.font.add_file('data/audiowide-regular.ttf')


class LoopLayer(Layer):
    def __init__(self, times_played=0):
        super(LoopLayer, self).__init__()

        self.is_playing = True
        self.times_played = times_played

        self.mushrooms = []
        self.flying = []
        self.bullets = []

        self.player = Player()
        self.player_layer = PlayerLayer(self.player, self.on_player_shoot)
        self.explosion_layer = ExplosionLayer()
        self.mushroom_layer = MushroomsLayer(self.player, self.on_mushroom_add)
        self.messages_layer = MessagesLayer(self.on_end)
        self.moon = Moon()

        self.add(BackgroundLayer(), z=0)
        self.add(self.player_layer, z=1)
        self.add(self.mushroom_layer, z=2)
        self.add(MoonLayer(self.moon), z=3)
        self.add(self.explosion_layer, z=5)
        self.add(self.messages_layer, z=6)

        self.cutting_manager = CollisionManagerGrid(
            -200, cfg.W + 200, -200, cfg.H + 200, 40, 40
        )
        self.shooting_manager = CollisionManagerGrid(
            -200, cfg.W + 200, -200, cfg.H + 200, 40, 40
        )
        self.schedule(self.update)

    def update(self, dt):
        if not self.is_playing:
            return

        self.cutting_manager.clear()
        self.shooting_manager.clear()

        self.update_objects(dt)
        self.update_physics()
        self.mark_for_releasing()
        self.mark_for_shooting()

        self.mushrooms = [m for m in self.mushrooms
                          if not m.is_flying]
        self.flying = [m for m in self.flying
                       if not m.is_shot and m.is_flying]
        self.bullets = [b for b in self.bullets
                        if not b.is_used]

        if len(self.mushrooms) < cfg.SHR_MAX_COUNT:
            self.mushroom_layer.start_generating()
        else:
            self.mushroom_layer.stop_generating()

    def update_objects(self, dt):
        self.player.update(dt)
        for mushroom in self.flying:
            mushroom.update(dt)

    def update_physics(self):
        self.player.update_shape()
        for mushroom in self.mushrooms:
            mushroom.update_shape()
            self.cutting_manager.add(mushroom)

        for mushroom in self.flying:
            mushroom.update_shape()
            self.shooting_manager.add(mushroom)

    def mark_for_releasing(self):
        for mushroom in self.cutting_manager.iter_colliding(self.player):
            self.flying.append(mushroom)
            sounds.play_cut()
            mushroom.start_flying()

    def mark_for_shooting(self):
        objs_touching_point = self.shooting_manager.objs_touching_point
        shot_mushrooms = set()
        for bullet in self.bullets:
            b_x, b_y = bullet.position
            if self.moon.cshape.touches_point(b_x, b_y):
                bullet.use()
                continue

            for mushroom in objs_touching_point(b_x, b_y):
                shot_mushrooms.add(mushroom)
                bullet.use()
                break  # One bullet -- one mushroom

        for mushroom in shot_mushrooms:
            if mushroom.mark_as_shot():
                sounds.play_poof()
                self.messages_layer.on_kill()
                self.explosion_layer.explode(mushroom.position)

    def on_mushroom_add(self, mushroom):
        self.mushrooms.append(mushroom)

    def on_player_shoot(self, gun_position, mouse_position):
        if not self.is_playing:
            return

        sounds.play_shot()
        for i in xrange(cfg.BULLET_COUNT):
            direction = mouse_position - gun_position
            angle = atan2(direction.x, direction.y)
            random_delta = (random() - 0.5) * cfg.BULLET_DIR_RANDOM
            angle = radians(degrees(angle) + random_delta)

            bullet = Bullet(gun_position, Vector2(sin(angle), cos(angle)))
            self.player.parent.add(bullet)
            self.bullets.append(bullet)

    def on_end(self, n_mushrooms):
        self.stop()
        if not self.is_playing:
            return

        if self.times_played == cfg.END_EXPLANATION_N:
            print cfg.END_EXPLANATION
        self.stop_playing(n_mushrooms)

    def stop_playing(self, n_mushrooms):
        self.is_playing = False
        self.mushroom_layer.stop()
        self.player.stop()
        self.player.hand.stop()
        self.player_layer.is_controlable = False

        self.add(TheEndLayer(
            n_mushrooms=n_mushrooms,
            scene_class=type(self.parent),
            times_played=self.times_played
        ), z=7)


class BackgroundLayer(Layer):
    def __init__(self):
        super(BackgroundLayer, self).__init__()
        self.add(Background(), z=0)
        self.add(StartsLayer(), z=1)
        self.add(Sun(), z=1)
        self.add(Earth(), z=1)


class StartsLayer(Layer):
    def __init__(self):
        super(StartsLayer, self).__init__()

        positions = []
        for i in xrange(cfg.STAR_COUNT):
            for j in xrange(cfg.STAR_MAX_ATTEMPT):
                new_pos = Vector2(random() * cfg.W, random() * cfg.H)
                is_ok = True
                for pos in positions:
                    if (pos - new_pos).magnitude() < cfg.STAR_MIN_DISTANCE:
                        is_ok = False
                        break
                if is_ok:
                    positions.append(new_pos)
                    break

        for pos in positions:
            self.add(Star(pos))


class MoonLayer(Layer):
    def __init__(self, moon):
        super(MoonLayer, self).__init__()
        self.add(moon)


class PlayerLayer(Layer):
    ccwise_keys = key.LEFT, key.A
    cwise_keys = key.RIGHT, key.D

    is_event_handler = True

    def __init__(self, player, on_player_shoot):
        super(PlayerLayer, self).__init__()
        self.is_controlable = True

        self.player = player
        self.add(self.player, z=0)
        self.add(self.player.hand, z=1)
        self.on_player_shoot = on_player_shoot

    def on_key_press(self, key, modifiers):
        if not self.is_controlable:
            return

        if key in self.ccwise_keys:
            self.player.start_moving_ccwise()
        elif key in self.cwise_keys:
            self.player.start_moving_cwise()

    def on_key_release(self, key, modifiers):
        if not self.is_controlable:
            return

        if key in self.ccwise_keys:
            self.player.stop_moving_ccwise()
        elif key in self.cwise_keys:
            self.player.stop_moving_cwise()

    def on_mouse_press(self, x, y, button, modifiers):
        if not self.is_controlable:
            return

        self.on_player_shoot(self.player.get_gun_pointer(), Vector2(x, y))

    def on_mouse_motion(self, x, y, dx, dy):
        self.player.on_mouse_motion(Vector2(x, y))


class MushroomsLayer(Layer):
    def __init__(self, player, on_mushroom_add):
        super(MushroomsLayer, self).__init__()
        self.player = player
        self.on_mushroom_add = on_mushroom_add

        self.is_generate = True
        self.generate_mushrooms()

    def delay_generate_mushrooms(self):
        tmin, tmax = cfg.SHR_GEN_TIMEOUT_MIN, cfg.SHR_GEN_TIMEOUT_MAX
        self.do(
            Delay(tmin + (tmax - tmin) * random()) +
            CallFunc(self.generate_mushrooms)
        )

    def generate_mushrooms(self, dt=None):
        if not self.is_generate:
            self.delay_generate_mushrooms()
            return

        pld = cfg.SHR_GEN_DISTANCE_FROM_PL
        rotation = (self.player.rotation + pld + (360.0 - 2 * pld) * random())
        rotation %= 360.0

        for i in cfg.SHR_GEN_RAND_DIST_INDEXES:
            threshhold = (cfg.SHR_GEN_THRESHOLD_MAIN +
                          cfg.SHR_GEN_THRESHOLD_REL * abs(i))
            if random() < threshhold:
                continue
            dst_rel = cfg.SHR_GEN_DISTANCE_REL
            mushroom = Mushroom(rotation + i * (dst_rel + random() * dst_rel))
            self.add(mushroom)
            self.on_mushroom_add(mushroom)

        self.delay_generate_mushrooms()

    def start_generating(self):
        self.is_generate = True

    def stop_generating(self):
        self.is_generate = False


class ExplosionLayer(Layer):
    def explode(self, position):
        self.add(Explosion(position))


class MessagesLayer(Layer):
    def __init__(self, on_end):
        super(MessagesLayer, self).__init__()
        self.on_end = on_end
        self.label = Label(
            'Mushrooms: 000',
            position=(cfg.W - cfg.PADDING,
                      cfg.H - cfg.PADDING),
            font_name='Audiowide',
            font_size=14,
            anchor_x='right',
            anchor_y='top'
        )
        self.add(self.label)
        self.label.visible = False
        self.n_mushrooms = 0

        self.generate_n_required()

    def on_kill(self):
        self.n_mushrooms += 1
        self.label.visible = True
        self.label.element.text = 'Mushrooms: {0:03d}'.format(self.n_mushrooms)

        if self.n_mushrooms >= self.n_mushrooms_required:
            self.label.visible = False
            self.on_end(self.n_mushrooms)

    def generate_n_required(self):
        # Yes. Yes. I know. This does not make sense (as the whole game).
        # Player wins if it shoots stupid number of mushrooms.
        # Each time it is different random number.
        #
        # By the wat, thanks for reading my source code. :)
        n = -1
        for i in xrange(20):
            n = randint(
                cfg.SHR_REQUIRED_MIN,
                cfg.SHR_REQUIRED_MAX
            )
            # Do not allow to nice numbers (10, 20, 25 etc.)
            if n % 10 != 0 and n % 5 != 0:
                break
        if n != -1:
            self.n_mushrooms_required = n
        else:
            self.n_mushrooms_required = 34  # Default stupid value


class AbstractInfoLayer(Layer):
    is_event_handler = True

    def __init__(self, scene_class, times_played):
        super(AbstractInfoLayer, self).__init__()

        self.scene_class = scene_class
        self.times_played = times_played

        self.add(ColorLayer(*cfg.BACK_COLOR), z=2)
        self.add(Label(
            self.get_play_msg_text(),
            position=(cfg.HALF_W, cfg.PADDING),
            font_name='Audiowide',
            font_size=12,
            anchor_x='center',
            anchor_y='bottom'
        ), z=3)

    def on_key_press(self, key, modifiers):
        if key == pyglet.window.key.SPACE:
            director.replace(self.scene_class(self.times_played + 1))


class TheEndLayer(AbstractInfoLayer):
    def __init__(self, n_mushrooms, scene_class, times_played):
        super(TheEndLayer, self).__init__(scene_class, times_played)

        # Yeah. Achievements are totally random too.
        # Thank You, for peaking around. :)
        text = cfg.END_CONG_TEXT_TEMPLATE.format(
            n_mushrooms, choice(cfg.END_ACHIEVEMENTS)
        )
        self.add(Label(
            text,
            position=cfg.CENTER,
            font_name='Audiowide',
            font_size=24,
            width=500,
            multiline=True,
            anchor_x='center',
            anchor_y='center'
        ), z=3)

    def get_play_msg_text(self):
        return 'press SPACE to play again'


class StartLayer(AbstractInfoLayer):
    def __init__(self, scene_class):
        super(StartLayer, self).__init__(scene_class, 0)
        self.add(Background(), z=0)
        self.add(StartsLayer(), z=1)

        self.add(Label(
            cfg.START_TEXT,
            position=(cfg.HALF_W, cfg.HALF_H + 100),
            font_name='Audiowide',
            font_size=24,
            width=500,
            multiline=True,
            anchor_x='center',
            anchor_y='center'
        ), z=3)

        self.add(Label(
            cfg.START_CONTROLS,
            position=(cfg.HALF_W, cfg.HALF_H - 100),
            font_name='Audiowide',
            font_size=16,
            width=500,
            multiline=True,
            anchor_x='center',
            anchor_y='center'
        ), z=3)

    def get_play_msg_text(self):
        return 'press SPACE to start'
