import math
from random import random, choice

import pyglet

from cocos.sprite import Sprite
from cocos.actions import (Repeat, RotateBy, ScaleTo, MoveBy, FadeOut,
                           CallFunc, Delay, Show)
from cocos.euclid import Vector2
from cocos.collision_model import CircleShape

from shrooms import cfg


player_still_img = pyglet.image.load('data/player-still.png')
player_still_grid = pyglet.image.ImageGrid(player_still_img, 1, 4)
player_still_anim = player_still_grid.get_animation(0.05)

player_still_r_img = pyglet.image.load('data/player-still-reverse.png')
player_still_r_grid = pyglet.image.ImageGrid(player_still_r_img, 1, 4)
player_still_r_anim = player_still_r_grid.get_animation(0.05)

player_move_img = pyglet.image.load('data/player-move.png')
player_move_grid = pyglet.image.ImageGrid(player_move_img, 1, 4)
player_move_anim = player_move_grid.get_animation(0.05)

player_move_r_img = pyglet.image.load('data/player-move-reverse.png')
player_move_r_grid = pyglet.image.ImageGrid(player_move_r_img, 1, 4)
player_move_r_anim = player_move_r_grid.get_animation(0.05)

player_hand_img = pyglet.image.load('data/player-hand.png')
player_hand_r_img = pyglet.image.load('data/player-hand-reverse.png')

background_img = pyglet.image.load('data/back.png')
earth_img = pyglet.image.load('data/earth.png')
sun_img = pyglet.image.load('data/sun.png')
star_img = pyglet.image.load('data/star.png')
moon_img = pyglet.image.load('data/moon.png')
mushroom_img = pyglet.image.load('data/mushroom.png')
cut_img = pyglet.image.load('data/cut.png')
bullet_img = pyglet.image.load('data/bullet.png')
explosion_imgs = [
    pyglet.image.load('data/explosion01.png'),
    pyglet.image.load('data/explosion02.png'),
    pyglet.image.load('data/explosion03.png'),
    pyglet.image.load('data/explosion04.png'),
    pyglet.image.load('data/explosion05.png'),
    pyglet.image.load('data/explosion06.png'),
    pyglet.image.load('data/explosion07.png'),
]


class Moon(Sprite):
    def __init__(self):
        super(Moon, self).__init__(
            moon_img,
            position=cfg.CENTER,
            scale=0.8,
        )
        self.cshape = CircleShape(cfg.CENTER, cfg.MOON_R)


class AroundMoonSprite(Sprite):
    def __init__(self, *args, **kwargs):
        super(AroundMoonSprite, self).__init__(*args, **kwargs)
        self.cshape = CircleShape(
            self.get_shape_pos(),
            self.get_shape_radius()
        )

    def update_shape(self):
        self.cshape.center = self.get_shape_pos()
        self.cshape.r = self.get_shape_radius()

    def update_pos_by_rot(self):
        angle = math.radians(self.rotation)
        delta = Vector2(math.sin(angle), math.cos(angle)) * cfg.MOON_R
        self.position = cfg.CENTER + delta


class Hand(Sprite):
    def __init__(self):
        super(Hand, self).__init__(
            player_hand_img,
            anchor=cfg.HAND_ANCHOR,
        )
        self.is_reverse = False

    def update_pos_and_rot(self, player, mouse_pos):
        self.update_pos(player)
        self.update_rot(mouse_pos)

    def update_pos(self, player):
        pl_angle = math.radians(player.rotation)
        direction = Vector2(math.sin(pl_angle), math.cos(pl_angle))
        length = cfg.HAND_POS.magnitude()
        real_pos = Vector2(*player.position) + length * direction
        self.position = real_pos

    def update_rot(self, mouse_pos):
        if mouse_pos is None:
            return
        x, y = (mouse_pos - self.position).normalized()

        angle = math.atan2(-y, x)
        if self.is_reverse:
            self.rotation = math.degrees(angle) + 180.0
        else:
            self.rotation = math.degrees(angle)

    def get_gun_pointer(self):
        if self.is_reverse:
            angle = math.radians(self.rotation - 90.0)
        else:
            angle = math.radians(self.rotation + 90.0)
        direction = Vector2(math.sin(angle), math.cos(angle))
        return Vector2(*self.position) + direction * cfg.HAND_GUN_DISTANCE

    def reverse(self, is_reverse):
        self.is_reverse = is_reverse
        if is_reverse:
            self.image = player_hand_r_img
            self.image_anchor = cfg.HAND_ANCHOR_R
        else:
            self.image = player_hand_img
            self.image_anchor = cfg.HAND_ANCHOR


class Player(AroundMoonSprite):
    def __init__(self):
        super(Player, self).__init__(
            player_still_anim,
            position=cfg.CENTER + (0, cfg.MOON_R),
            anchor=cfg.PL_ANCHOR
        )
        self.hand = Hand()
        self.mouse_pos = Vector2(cfg.W, cfg.H)

        self.is_reverse = False

        self.ccwise_action = None
        self.cwise_action = None

    def on_mouse_motion(self, mouse_pos):
        self.mouse_pos = mouse_pos
        self.update_images()

    def update(self, dt):
        self.update_pos_by_rot()
        self.hand.update_pos_and_rot(self, self.mouse_pos)
        self.update_images()

    def get_shape_pos(self):
        return Vector2(*self.position)

    def get_shape_radius(self):
        return cfg.PL_SHAPE_RADIUS

    def get_gun_pointer(self):
        return self.hand.get_gun_pointer()

    def start_moving_cwise(self):
        self.stop_moving_actions()
        self.cwise_action = self.do(Repeat(
            RotateBy(cfg.PL_SPEED, cfg.DEFAULT_T)
        ))
        self.update_images()

    def start_moving_ccwise(self):
        self.stop_moving_actions()
        self.ccwise_action = self.do(Repeat(
            RotateBy(-cfg.PL_SPEED, cfg.DEFAULT_T)
        ))
        self.update_images()

    def stop_moving_cwise(self):
        if self.cwise_action:
            self.remove_action(self.cwise_action)
        self.cwise_action = None
        self.update_images()

    def stop_moving_ccwise(self):
        if self.ccwise_action:
            self.remove_action(self.ccwise_action)
        self.ccwise_action = None
        self.update_images()

    def stop_moving_actions(self):
        self.stop_moving_cwise()
        self.stop_moving_ccwise()

    def update_images(self):
        hand_r = self.get_relative_hand_rot()

        if self.is_reverse and abs(hand_r) > 90.0:
            self.reverse(False)
        elif not self.is_reverse and abs(hand_r) > 90.0:
            self.reverse(True)
        else:
            # Stupid hack. To update flying VS still images
            self.reverse(self.is_reverse)

    def get_relative_hand_rot(self):
        delta_r = self.hand.rotation - self.rotation
        if delta_r > 180.0:
            delta_r -= 360.0
        if delta_r < -180.0:
            delta_r += 360.0
        return delta_r

    def reverse(self, is_reverse):
        self.is_reverse = is_reverse

        if self.is_reverse:
            if self.is_still:
                new_image = player_still_r_anim
            else:
                new_image = player_move_r_anim
            self.image_anchor = cfg.PL_ANCHOR_R
            self.hand.reverse(self.is_reverse)
        else:
            if self.is_still:
                new_image = player_still_anim
            else:
                new_image = player_move_anim
            self.image_anchor = cfg.PL_ANCHOR
            self.hand.reverse(self.is_reverse)

        if new_image is not self.image:
            self.image = new_image

    @property
    def is_still(self):
        return self.ccwise_action is None and self.cwise_action is None


class Mushroom(AroundMoonSprite):
    def __init__(self, angle):
        super(Mushroom, self).__init__(
            mushroom_img,
            rotation=angle,
            scale=cfg.SHR_INIT_SCALE,
            anchor=cfg.SHR_ANCHOR
        )
        self.update_pos_by_rot()
        self.start_growing()

        self.is_flying = False
        self.is_shot = False

    def update(self, dt):
        if self.is_flying:
            x, y = self.position - cfg.CENTER
            if abs(x) > cfg.HALF_W or abs(y) > cfg.HALF_H:
                self.start_dissapearing()

    def start_growing(self):
        scale_to = (cfg.SHR_SCALE_MIN +
                    (cfg.SHR_SCALE_MAX - cfg.SHR_SCALE_MIN) * random())
        speed = (cfg.SHR_SCALE_SPEED_MIN +
                 (cfg.SHR_SCALE_SPEED_MAX - cfg.SHR_SCALE_SPEED_MIN) *
                 random())
        self.do(Repeat(ScaleTo(scale_to, speed)))

    def start_flying(self):
        self.is_flying = True
        self.stop()  # Stops growing
        move_by = self.get_flying_direction() * cfg.SHR_FLY_SPEED
        self.do(
            Delay(cfg.SHR_FLY_DELAY) + Repeat(MoveBy(move_by, cfg.DEFAULT_T))
        )

        self.generate_cut()

    def start_dissapearing(self):
        self.is_flying = False
        self.stop()
        self.kill()

    def mark_as_shot(self):
        if self.is_flying:
            self.is_shot = True
            self.kill()
            return True
        return False

    def get_flying_direction(self):
        angle = math.radians(self.rotation)
        return Vector2(math.sin(angle), math.cos(angle))

    def get_shape_pos(self):
        return Vector2(*self.position)

    def get_shape_radius(self):
        return cfg.SHR_SHAPE_RADIUS * self.scale

    def generate_cut(self):
        angle = math.radians(self.rotation)
        direction = Vector2(math.sin(angle), math.cos(angle))
        length = cfg.CUT_POS_DELTA

        cut = Sprite(
            cut_img,
            position=Vector2(*self.position) + length * direction,
            scale=self.scale + cfg.CUT_SCALE_DELTA,
            rotation=self.rotation
        )

        self.parent.add(cut)
        cut.visible = False
        cut.do(Show() + FadeOut(cfg.CUT_FADE_OUT_TIMEOUT) + CallFunc(cut.kill))


class Bullet(Sprite):
    def __init__(self, position, direction):
        super(Bullet, self).__init__(
            bullet_img,
            position=position
        )
        self.is_used = False
        speed = cfg.BULLET_SPEED + (random() - 0.5) * cfg.BULLET_SPEED_RANDOM
        self.do(
            MoveBy(direction * cfg.MAX_DISTANCE, speed) + CallFunc(self.use)
        )

    def use(self):
        self.is_used = True
        self.stop()
        self.kill()


class Explosion(Sprite):
    def __init__(self, position):
        super(Explosion, self).__init__(
            choice(explosion_imgs),
            #scale=0.8,
            position=position,
            rotation=360.0 * random()
        )
        self.do(
            Delay(cfg.EXPLOSION_INIT_DELAY) +
            FadeOut(cfg.EXPLOSION_FADE_OUT_TIME) +
            CallFunc(self.kill)
        )


class Background(Sprite):
    def __init__(self):
        super(Background, self).__init__(background_img, anchor=(0, 0))


class MovingAsEllispe(Sprite):
    def __init__(self):
        super(MovingAsEllispe,  self).__init__(
            self.get_img(),
            position=self.get_init_pos(),
            rotation=self.get_init_rot()
        )

        self.t = self.get_init_t()
        self.schedule(self.update)
        self.a, self.b = self.get_ab()
        self.cycle_t = self.get_cycle_t()
        self.update(0.0)

    def update(self, dt):
        self.t += dt / self.cycle_t
        if self.t > 1.0:
            self.t = 0.0

        angle = math.radians(360.0 * self.t)
        x, y = math.sin(angle) * 150, math.cos(angle) * 200
        self.image_anchor = (x, y)


class Earth(MovingAsEllispe):
    def get_init_rot(self):
        return cfg.EARTH_ROT

    def get_init_pos(self):
        return cfg.EARTH_POS

    def get_img(self):
        return earth_img

    def get_ab(self):
        return cfg.EARTH_A, cfg.EARTH_B

    def get_cycle_t(self):
        return cfg.EARTH_CYCLE_T

    def get_init_t(self):
        return cfg.EARTH_INIT_T


class Sun(MovingAsEllispe):
    def get_init_rot(self):
        return cfg.SUN_ROT

    def get_init_pos(self):
        return cfg.SUN_POS

    def get_img(self):
        return sun_img

    def get_ab(self):
        return cfg.SUN_A, cfg.SUN_B

    def get_cycle_t(self):
        return cfg.SUN_CYCLE_T

    def get_init_t(self):
        return cfg.SUN_INIT_T


class Star(Sprite):
    def __init__(self, position):
        super(Star, self).__init__(
            star_img,
            position=position,
        )
        self.delay_blink()

    def delay_blink(self):
        self.do(
            Delay(random() * cfg.STAR_BLINK_DELAY) +
            CallFunc(self.blink)
        )

    def blink(self):
        self.do(
            ScaleTo(cfg.STAR_BLINK_SCALE, cfg.STAR_BLINK_T) +
            ScaleTo(1.0, cfg.STAR_BLINK_T)
        )
        self.delay_blink()
